<?php
/**
 * Database Exporter
 * The exporter creates an overview of the complete database 
 * and analyzes the structure with common rules by RoR.
 * See http://rails.raaum.org/database-conventions.html for 
 * more details.
 * 
 * Actually, the structure analyzer works at the moment
 * <li>only with PHP/MySql
 * <li>is only based on names, not InnoDB Engines's Foreign Keys
 * <li>therefor the "finding foreign keys" HAVE TO be unique
 * <li>additionally, all fieldtypes (of key and his foreign keys) 
 * 	are compared
 * 
 * @author Jan Philipp <knallisworld@googlemail.com>
 * @version 1.0
 */

/* // example
 
// load class
require_once('database_exporter.php');

// init objectinstance
$dbexporter = new Database_Exporter(array(
	'server' => 'SERVER',
	'username' => 'USERNAME',
	'password' => 'PASSWORD'
));

if ($dbexporter) {
	// export 'DATABASENAME' to subdirectory xml/
	if ($dbexporter->exportDbStructureToXML('DATABASENAME', 'xml')) {
		$dbexporter->close();
	
		echo sprintf(
			'There were %d XML files generated in directory: <a href="%s">%s</a>', 
			$dbexporter->getCountedFilesWritten(), 
			'xml/database.xml', 
			'xml');
	}
	else {
		echo 'error in database';
	}
}
else {
	echo 'no database!';
}


 */

// linefeed
define('LF', "\n");

/**
 * Database Exporter
 * The exporter creates an overview of the complete database 
 * and analyzes the structure with common rules by RoR.
 * 
 * @author Jan Philipp <knallisworld@googlemail.com>
 * @version 1.0
 * @use mysqli
 */
class Database_Exporter {
	
	/**
	 * Connection to a database
	 *
	 * @var mysqli $connection
	 */
	private $connection = null;
	
	/**
	 * Loaded table names
	 *
	 * @var array $tableNames
	 */
	private $tableNames = array();

	/**
	 * Loaded table data
	 *
	 * @var array $tableData
	 */
	private $tableData = array();
	
	/**
	 * references
	 * 
	 * @var array $references
	 */
	private $references = array();
	
	/**
	 * Used database name
	 *
	 * @var string $$databaseName
	 */
	private $databaseName = array();
	
	/**
	 * temporary table-fields
	 * 
	 * @var array $tableFields
	 */
	private $tableFields = array();
	
	/**
	 * how many files wrote
	 * 
	 * @var int $countFilesWritten
	 */
	private $countFilesWritten = 0;
	
	private $xml = '';
	private $prolog = '';
	private $created = '';
	
	public function __construct($confData) {
		$this->connection = new mysqli($confData['server'], $confData['username'], $confData['password']);
		if (!$this->connection) {
			return false;
		}
		
		$this->prolog = '<?xml version="1.0"?>' . LF . '<?xml-stylesheet href="exported_database.xsl" type="text/xsl" ?> ' . LF;
		$this->created = sprintf('<created timestamp="%s">%s</created>', time(), date('d/m/y H:i:s')) . LF;
		return true;
	}
	
	private function init() {
		$this->tableNames 			= array();
		$this->tableData 			= array();
		$this->references 			= array();
		$this->databaseName 		= array();
		$this->tableFields 			= array();
		$this->countFilesWritten 	= 0;
		$this->xml 					= '';
	}
	
	/**
	 * Loading the current structure of the given databse and store the informations locally.
	 *
	 * @param $databaseName
	 * @return boolean true if successfully
	 */
	private function loadDatabaseStructure($databaseName = null) {
		$this->init();
		
		if ($databaseName == null || !$this->connection || $this->connection == null){
			echo 'no server connection or given database name';
			return false;
		}

		if (!$this->connection->select_db($databaseName)) {
			echo 'no database found by given name';
			return false;
		}
		$this->databaseName = $databaseName;
		
		$this->getTablenames();
		foreach ($this->tableNames as $tableName) {
			$this->loadTableData($tableName);
		}
		
		return true;
	}
	
	private function getTablenames() {
		$rows = array();
		if ($resultSet = $this->connection->query("SHOW TABLES;")) {
			while ($row = $resultSet->fetch_row()) {
				$rows[] = $row[0];
			}
			$resultSet->close();
		}
		$this->tableNames = $rows;
	}
	
	private function loadTableData($tableName = '') {
		if($tableName == '') {
			return null;
		}
		
		$rows = array();
		if ($resultSet = $this->connection->query(sprintf("SHOW FIELDS FROM %s;", $tableName))) {
			while ($row = $resultSet->fetch_assoc()) {
				$rows[] = $row;
			}
			$resultSet->close();
		}
		
		$this->tableData[$tableName] = $rows;
	}
	
	/**
	 * exports the current database structure of $databaseName to $directory
	 * creates following files
	 * <li>structure.xml 	- an overview and checked relations
	 * <li>database.xml 	- an overview of all tables with linkings of TABLENAME.xml
	 * <li>TABLENAME.xml	- an overview of one table (for each found table)
	 * 
	 * @var string $databaseName
	 * @var string $directory (default '' == current)
	 */
	public function exportDbStructureToXML($databaseName, $directory = '') {
		if (!$this->loadDatabaseStructure($databaseName)) {
			return false;
		}
		
		foreach ($this->tableData as $tableName => $tableData) {
			$xml = sprintf('<table name="%s">', $tableName) . LF;
			$xml .= $this->created;
			$primaryKey = '';
			$indexKeys = array();
			foreach ($tableData as $tableField) {
				$xml .=sprintf('<field name="%s" autoincrement="%s">', $tableField['Field'], ($tableField['Extra'] == 'auto_increment') ? 'yes' : 'no') . LF;
				$type = explode(' ', $tableField['Type']);
				$xml .=sprintf('<type name="%s" unsigned="%s"/>', $type[0], (count($type) > 1) ? 'yes' : 'no') . LF;
				if ($tableField['Default'] != '') {
					$xml .=sprintf('<default>%s</default>', $tableField['Default']) . LF;
				}
				$xml .=sprintf('</field>') . LF;
				if ($tableField['Key'] == 'PRI') {
					$primaryKey = $tableField['Field'];
				}
				if ($tableField['Key'] == 'MUL') {
					$indexKeys[] = $tableField['Field'];
				}
			}
			
			$xml .='<constraints>' . LF;
			if ($primaryKey != '') {
				$xml .=sprintf('<field type="primary">%s</field>', $primaryKey) . LF;
			}
			foreach ($indexKeys as $indexKey) {
				$xml .=sprintf('<field type="index">%s</field>', $indexKey) . LF;
			}
			$xml .='</constraints>' . LF;
			$xml .= '</table>';
			
			// write one table overview with fields etc
			$this->writeFile($directory . '/' . $tableName . '.xml', $this->prolog . $xml);
			$this->xml .= $xml;
		}
		
		$xml = sprintf('<database name="%s">', $this->databaseName) . LF;
		$xml .= $this->created;
		foreach ($this->tableNames as $tableName) {
			$xml .=sprintf('<table>%s</table>', $tableName) . LF;
		}
		
		// retrive references
		$this->findReferences();
		$xml .= '<references>' . LF;
		foreach ($this->references as $field => $data) {
			$xml .= sprintf('<field name="%s">', $field) . LF;
			$xml .= sprintf('<table type="master">%s</table>', $data['master']) . LF;
			foreach ($data['details'] as $detailTable) {
				$xml .= sprintf('<table type="detail">%s</table>', $detailTable) . LF;
			}
			$xml .= '</field>';
		}
		$xml .= '</references>' .LF;
		$xml.='</database>';
		
		// write global tables overview = database
		$this->writeFile($directory . '/database.xml', $this->prolog . $xml);
		$this->xml = $xml . $this->xml;
		
		$xml = '<structure>' . LF;
		$xml .= $this->created;
		$xml .= '<messages>' . LF;
		$message = array();
		foreach (array_unique($this->messages) as $msg) {
			list($message['table'], $message['field'], $message['text']) = explode(';', $msg);
			$xml .= sprintf('<message table="%s" field="%s">', $message['table'], $message['field']) . LF;
			$xml .= sprintf('<text>%s</text>', $message['text']) . LF;
			$xml .= '</message>' . LF;
		}
		$xml .= '</messages>' . LF;
		$xml .= $this->xml;
		$xml .= '</structure>' . LF;
		
		// write complete file
		$this->writeFile($directory . '/structure.xml', $this->prolog . $xml);
		
		// copy if not current directory
		if ($directory != '') {
			@copy('exported_database.xsl', $directory . '/exported_database.xsl');
			@copy('default.css', $directory . '/default.css');
		}
		
		return true;
	}
	
	public function close() {
		$this->connection->close();
	}
	
	private function writeFile($destination, $content) {
		if (FALSE !== strpos ($destination, '/')) {
			$items = explode('/', $destination);
			
			$path = '';
			for ($i = 0, $c = count($items); $i < ($c - 1); $i++) {
				$path .= $items[$i] . '/';
				if (!file_exists($path)) {
					mkdir($path);
				}
			}
			$path .= $items[$i] ;
		}
		else {
			$path = $destination;
		}
		
		file_put_contents($path, $content);
		$this->countFilesWritten++;
	}
	
	public function getCountedFilesWritten() {
		return $this->countFilesWritten;
	}
	
	private function findReferences() {
		$this->copyFieldsToTemp();
		
		foreach ($this->tableData as $tableName => $tableData) {
			foreach ($tableData as $field) {
				if ($field['Key'] == 'PRI') {
					$tables = $this->getForeignKeys($field['Field'], $tableName, $field['Type']);
					$this->references[$field['Field']] = array(
						'master' => $tableName,
						'details' => $tables
					);
				}
			}
		}
	}
	
	private function copyFieldsToTemp() {
		foreach ($this->tableData as $tableName => $tableData) {
			$fields = array();
			foreach ($tableData as $field) {
				$fields[] = $field['Field'];
			}
			$this->tableFields[$tableName] = $fields;
		}
	}
	
	/**
	 * finds possible foreign keys (sub keys) in other tables except $exceptTable
	 */
	private function getForeignKeys($findField, $exceptTable, $findType) {
		$tables = array();
		foreach ($this->tableData as $tableName => $tableData) {
			if ($exceptTable != $tableName) {
				foreach ($tableData as $field) {
					if ($findField == $field['Field']) {
						$tables[] = $tableName;
						if ($field['Key'] == '') {
							$this->addMessage($tableName, $findField, 'no index (would be better, because it seems to be an index key)');
						}
						if ($field['Type'] <> $findType) {
							$this->addMessage($tableName, $findField, str_replace(';', '', sprintf("incorrect type (it is <code>%s</code>, but the foreign key's type is <code>%s</code>)", $field['Type'], $findType)));
						}
						continue;
					}
				}
			}
		}
		return $tables;
	}
	
	private $messages = array();
	private function addMessage($table, $field, $text) {
		$this->messages[] = sprintf('%s;%s;%s', $table, $field, $text);
	}
}
?>